use std::{
    fs::File,
    path::{Path, PathBuf},
};

use byte_unit::Byte;
use directories::ProjectDirs;
use log::trace;
use serde::{Deserialize, Serialize};

use crate::tildepathbuf::TildePathBuf;

const QUAL: &str = "liw.fi";
const ORG: &str = "Ambient CI";
const APP: &str = env!("CARGO_PKG_NAME");

const DEFAULT_CPUS: usize = 1;
const DEFAULT_MEMORY: Byte = Byte::GIBIBYTE;

#[derive(Debug, Deserialize)]
#[serde(deny_unknown_fields, default)]
struct StoredQemuConfig {
    cpus: Option<usize>,
    kvm_binary: Option<TildePathBuf>,
    memory: Option<Byte>,
    ovmf_vars_file: Option<TildePathBuf>,
    ovmf_code_file: Option<TildePathBuf>,
}

impl Default for StoredQemuConfig {
    fn default() -> Self {
        Self {
            cpus: None,
            memory: None,
            kvm_binary: Some(TildePathBuf::new("/usr/bin/kvm".into())),
            ovmf_vars_file: Some(TildePathBuf::new("/usr/share/ovmf/OVMF.fd".into())),
            ovmf_code_file: Some(TildePathBuf::new("/usr/share/ovmf/OVMF.fd".into())),
        }
    }
}

#[derive(Debug, Deserialize)]
#[serde(deny_unknown_fields, default)]
struct StoredConfig {
    tmpdir: Option<TildePathBuf>,
    image_store: Option<TildePathBuf>,
    projects: Option<TildePathBuf>,
    state: Option<TildePathBuf>,
    target: Option<String>,
    dput_target: Option<String>,
    executor: Option<TildePathBuf>,
    artifacts_max_size: Option<Byte>,
    cache_max_size: Option<Byte>,
    qemu: StoredQemuConfig,
    cpus: Option<usize>,
    memory: Option<Byte>,
}

impl Default for StoredConfig {
    fn default() -> Self {
        let dirs = ProjectDirs::from(QUAL, ORG, APP).expect("have home directory");
        #[allow(clippy::unwrap_used)]
        let state = dirs.state_dir().unwrap();

        Self {
            tmpdir: Some(PathBuf::from("/tmp").into()),
            image_store: Some(TildePathBuf::new(state.join("images"))),
            projects: Some(dirs.config_dir().join("projects.yaml").into()),
            state: Some(TildePathBuf::new(state.join("projects"))),
            target: None,
            dput_target: None,
            executor: None,
            qemu: Default::default(),
            artifacts_max_size: Byte::MEBIBYTE.multiply(10),
            cache_max_size: Byte::GIBIBYTE.multiply(10),
            cpus: None,
            memory: None,
        }
    }
}

#[derive(Debug, Serialize, Deserialize)]
#[serde(deny_unknown_fields)]
pub struct QemuConfig {
    cpus: usize,
    memory: Byte,
    kvm_binary: PathBuf,
    ovmf_vars_file: PathBuf,
    ovmf_code_file: PathBuf,
}

#[derive(Debug, Serialize, Deserialize)]
#[serde(deny_unknown_fields)]
pub struct Config {
    tmpdir: PathBuf,
    image_store: PathBuf,
    projects: PathBuf,
    state: PathBuf,
    target: Option<String>,
    dput_target: Option<String>,
    executor: Option<PathBuf>,
    artifacts_max_size: Byte,
    cache_max_size: Byte,
    qemu: QemuConfig,
}

impl Config {
    pub fn new() -> Result<Self, ConfigError> {
        Self::try_from(StoredConfig::default())
    }

    pub fn from_yaml_file(filename: &Path) -> Result<Self, ConfigError> {
        let f = File::open(filename).map_err(|e| ConfigError::Read(filename.into(), e))?;
        let stored: StoredConfig =
            serde_yml::from_reader(f).map_err(|e| ConfigError::Yaml(filename.into(), e))?;
        let config = Config::try_from(stored)?;
        trace!(
            "Loaded configuration from file {}:\n{config:#?}",
            filename.display()
        );
        Ok(config)
    }

    pub fn tmpdir(&self) -> &Path {
        &self.tmpdir
    }

    pub fn image_store(&self) -> &Path {
        &self.image_store
    }

    pub fn projects(&self) -> &Path {
        &self.projects
    }

    pub fn state(&self) -> &Path {
        &self.state
    }

    pub fn target(&self) -> Option<&str> {
        self.target.as_deref()
    }

    pub fn dput_target(&self) -> Option<&str> {
        self.dput_target.as_deref()
    }

    pub fn executor(&self) -> Option<&Path> {
        self.executor.as_deref()
    }

    pub fn cpus(&self) -> usize {
        self.qemu.cpus
    }

    pub fn memory(&self) -> Byte {
        self.qemu.memory
    }

    pub fn kvm_binary(&self) -> PathBuf {
        self.qemu.kvm_binary.clone()
    }

    pub fn ovmf_vars_file(&self) -> PathBuf {
        self.qemu.ovmf_vars_file.clone()
    }

    pub fn ovmf_code_file(&self) -> PathBuf {
        self.qemu.ovmf_code_file.clone()
    }

    pub fn artifacts_max_size(&self) -> u64 {
        self.artifacts_max_size.as_u64()
    }

    pub fn cache_max_size(&self) -> u64 {
        self.cache_max_size.as_u64()
    }
}

impl TryFrom<StoredConfig> for Config {
    type Error = ConfigError;
    fn try_from(stored: StoredConfig) -> Result<Self, Self::Error> {
        if stored.cpus.is_some() {
            eprintln!("deprecated: the `cpus` field is replaced by `qemu.cpus`");
        }
        if stored.memory.is_some() {
            eprintln!("deprecated: the `memory` field is replaced by `qemu.memory`");
        }

        let qemu = QemuConfig {
            cpus: if let Some(cpus) = stored.qemu.cpus {
                cpus
            } else if let Some(cpus) = stored.cpus {
                cpus
            } else {
                DEFAULT_CPUS
            },
            memory: if let Some(memory) = stored.qemu.memory {
                memory
            } else if let Some(memory) = stored.memory {
                memory
            } else {
                DEFAULT_MEMORY
            },
            kvm_binary: stored
                .qemu
                .kvm_binary
                .ok_or(ConfigError::Missing("qemu.kvm_binary"))?
                .path()
                .into(),
            ovmf_vars_file: stored
                .qemu
                .ovmf_vars_file
                .ok_or(ConfigError::Missing("qemu.ovmf_vars_file"))?
                .path()
                .into(),
            ovmf_code_file: stored
                .qemu
                .ovmf_code_file
                .ok_or(ConfigError::Missing("qemu.ovmf_code_file"))?
                .path()
                .into(),
        };

        Ok(Self {
            tmpdir: stored
                .tmpdir
                .ok_or(ConfigError::Missing("tmpdir"))?
                .path()
                .into(),
            image_store: stored
                .image_store
                .ok_or(ConfigError::Missing("image_store"))?
                .path()
                .into(),
            projects: stored
                .projects
                .ok_or(ConfigError::Missing("projects"))?
                .path()
                .into(),
            state: stored
                .state
                .ok_or(ConfigError::Missing("state"))?
                .path()
                .into(),
            target: stored.target,
            dput_target: stored.dput_target,
            executor: stored.executor.map(|path| path.path().into()),
            artifacts_max_size: stored
                .artifacts_max_size
                .ok_or(ConfigError::Missing("artifacts_max_size"))?,
            cache_max_size: stored
                .cache_max_size
                .ok_or(ConfigError::Missing("cache_max_size"))?,
            qemu,
        })
    }
}

#[derive(Debug, thiserror::Error)]
pub enum ConfigError {
    #[error("failed to find home directory, while looking for configuration file")]
    ProjectDirs,

    #[error("failed to read configuration file {0}")]
    Read(PathBuf, #[source] std::io::Error),

    #[error("failed to parse configuration file as YAML: {0}")]
    Yaml(PathBuf, #[source] serde_yml::Error),

    #[error("programming error: stored config field {0} is missing")]
    Missing(&'static str),
}
